@@ -1,16 +1,3 @@
-/******************************************************************************
-*******************************************************************************
-**
-**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
-**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
-**
-**  This copyrighted material is made available to anyone wishing to use,
-**  modify, copy, or redistribute it subject to the terms and conditions
-**  of the GNU General Public License v.2.
-**
-*******************************************************************************
-******************************************************************************/
-
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -26,8 +13,7 @@
 #include "linux_endian.h"
 
 #define WANT_GFS_CONVERSION_FUNCTIONS
-#include <linux/gfs_ondisk.h>
-
+#include "gfs_ondisk.h"
 #include "gfshex.h"
 
 extern int line;
@@ -73,7 +59,9 @@ void do_dinode_extended(struct gfs_dinode *di, char *buf)
 	else if (di->di_type == GFS_FILE_DIR &&
 			 !(di->di_flags & GFS_DIF_EXHASH)) {
 		/* Directory Entries: */
-		for (x = sizeof(struct gfs_dinode); x < bufsize; x += de.de_rec_len) {
+		for (x = sizeof(struct gfs_dinode); x < bufsize &&
+				 de.de_rec_len > sizeof(struct gfs_dirent) &&
+				 de.de_rec_len < 256; x += de.de_rec_len) {
 			gfs_dirent_in(&de, buf + x);
 			if (de.de_inum.no_formal_ino)
 				indirect_block[indirect_blocks++] = de.de_inum.no_formal_ino;
