@@ -27,13 +27,15 @@
 #include "libgfs2.h"
 #include "gfs2hex.h"
 
+#include <syslog.h>
+
 #define RGLIST_DUMMY_BLOCK -2
 
 int display(int identify_only);
 extern void eol(int col);
 extern void do_leaf_extended(char *buf, struct iinfo *indir);
 extern int do_indirect_extended(char *buf, struct iinfo *ii);
-extern void savemeta(char *out_fn, int slow);
+extern void savemeta(const char *out_fn, int slow);
 extern void restoremeta(const char *in_fn, const char *out_device,
 			int printblocksonly);
 
@@ -147,7 +149,6 @@ int bobgets(char string[],int x,int y,int sz,int *ch)
 			case('\r'):
 				rc=1;
 				done=TRUE;
-				string[runningy-y] = '\0';
 				break;
 			case(KEY_CANCEL):
 			case(0x01B):
@@ -627,10 +628,10 @@ void rgcount(void)
 		block = sbd1->sb_rindex_di.no_addr;
 	else
 		block = masterblock("rindex");
-	ribh = bread(&sbd.buf_list, block);
+	ribh = bread(&sbd, block);
 	riinode = inode_get(&sbd, ribh);
 	printf("%lld RGs in this file system.\n",
-	       (unsigned long long)riinode->i_di.di_size / risize());
+	       riinode->i_di.di_size / risize());
 	inode_put(riinode, not_updated);
 	exit(EXIT_SUCCESS);
 }
@@ -710,30 +711,6 @@ void gfs_rgrp_print(struct gfs_rgrp *rg)
 	pv(rg, rg_freemeta, "%u", "0x%x");
 }
 
-/* ------------------------------------------------------------------------ */
-/* get_rg_addr                                                              */
-/* ------------------------------------------------------------------------ */
-uint64_t get_rg_addr(int rgnum)
-{
-	struct gfs2_buffer_head *bh;
-	uint64_t rgblk = 0, block;
-	struct gfs2_inode *riinode;
-
-	if (gfs1)
-		block = sbd1->sb_rindex_di.no_addr;
-	else
-		block = masterblock("rindex");
-	bh = bread(&sbd.buf_list, block);
-	riinode = inode_get(&sbd, bh);
-	if (rgnum < riinode->i_di.di_size / risize())
-		rgblk = find_rgrp_block(riinode, rgnum);
-	else
-		fprintf(stderr, "Error: File system only has %lld RGs.\n",
-			(unsigned long long)riinode->i_di.di_size / risize());
-	inode_put(riinode, not_updated);
-	return rgblk;
-}
-
 /* ------------------------------------------------------------------------ */
 /* set_rgrp_flags - Set an rgrp's flags to a given value                    */
 /* rgnum: which rg to print or modify flags for (0 - X)                     */
@@ -747,11 +724,25 @@ void set_rgrp_flags(int rgnum, uint32_t new_flags, int modify, int full)
 		struct gfs2_rgrp rg2;
 		struct gfs_rgrp rg1;
 	} rg;
-	struct gfs2_buffer_head *bh;
-	uint64_t rgblk;
+	struct gfs2_buffer_head *bh, *ribh;
+	uint64_t rgblk, block;
+	struct gfs2_inode *riinode;
 
-	rgblk = get_rg_addr(rgnum);
-	bh = bread(&sbd.buf_list, rgblk);
+	if (gfs1)
+		block = sbd1->sb_rindex_di.no_addr;
+	else
+		block = masterblock("rindex");
+	ribh = bread(&sbd, block);
+	riinode = inode_get(&sbd, ribh);
+	if (rgnum >= riinode->i_di.di_size / risize()) {
+		fprintf(stderr, "Error: File system only has %lld RGs.\n",
+			riinode->i_di.di_size / risize());
+		inode_put(riinode, not_updated);
+		brelse(ribh, not_updated);
+		return;
+	}
+	rgblk = find_rgrp_block(riinode, rgnum);
+	bh = bread(&sbd, rgblk);
 	if (gfs1)
 		gfs_rgrp_in(&rg.rg1, bh->b_data);
 	else
@@ -782,8 +773,9 @@ void set_rgrp_flags(int rgnum, uint32_t new_flags, int modify, int full)
 			       (unsigned long long)rgblk, rg.rg2.rg_flags);
 		brelse(bh, not_updated);
 	}
+	inode_put(riinode, not_updated);
 	if (modify)
-		bsync(&sbd.buf_list);
+		bsync(&sbd);
 }
 
 /* ------------------------------------------------------------------------ */
@@ -846,7 +838,7 @@ int parse_rindex(struct gfs2_inode *di, int print_rindex)
 			else {
 				struct gfs2_buffer_head *tmp_bh;
 
-				tmp_bh = bread(&sbd.nvbuf_list, ri.ri_addr);
+				tmp_bh = bread(&sbd, ri.ri_addr);
 				if (gfs1) {
 					struct gfs_rgrp rg1;
 					gfs_rgrp_in(&rg1, tmp_bh->b_data);
@@ -1386,7 +1378,7 @@ int display_extended(void)
 
 	/* Display any indirect pointers that we have. */
 	if (block_is_rindex()) {
-		tmp_bh = bread(&sbd.buf_list, block);
+		tmp_bh = bread(&sbd, block);
 		tmp_inode = inode_get(&sbd, tmp_bh);
 		parse_rindex(tmp_inode, TRUE);
 		brelse(tmp_bh, not_updated);
@@ -1398,34 +1390,33 @@ int display_extended(void)
 		return -1;
 	else if (block_is_rglist()) {
 		if (gfs1)
-			tmp_bh = bread(&sbd.buf_list,
-				       sbd1->sb_rindex_di.no_addr);
+			tmp_bh = bread(&sbd, sbd1->sb_rindex_di.no_addr);
 		else
-			tmp_bh = bread(&sbd.buf_list, masterblock("rindex"));
+			tmp_bh = bread(&sbd, masterblock("rindex"));
 		tmp_inode = inode_get(&sbd, tmp_bh);
 		parse_rindex(tmp_inode, FALSE);
 		brelse(tmp_bh, not_updated);
 	}
 	else if (block_is_jindex()) {
-		tmp_bh = bread(&sbd.buf_list, block);
+		tmp_bh = bread(&sbd, block);
 		tmp_inode = inode_get(&sbd, tmp_bh);
 		print_jindex(tmp_inode);
 		brelse(tmp_bh, not_updated);
 	}
 	else if (block_is_inum_file()) {
-		tmp_bh = bread(&sbd.buf_list, block);
+		tmp_bh = bread(&sbd, block);
 		tmp_inode = inode_get(&sbd, tmp_bh);
 		print_inum(tmp_inode);
 		brelse(tmp_bh, not_updated);
 	}
 	else if (block_is_statfs_file()) {
-		tmp_bh = bread(&sbd.buf_list, block);
+		tmp_bh = bread(&sbd, block);
 		tmp_inode = inode_get(&sbd, tmp_bh);
 		print_statfs(tmp_inode);
 		brelse(tmp_bh, not_updated);
 	}
 	else if (block_is_quota_file()) {
-		tmp_bh = bread(&sbd.buf_list, block);
+		tmp_bh = bread(&sbd, block);
 		tmp_inode = inode_get(&sbd, tmp_bh);
 		print_quota(tmp_inode);
 		brelse(tmp_bh, not_updated);
@@ -1438,6 +1429,8 @@ int display_extended(void)
 /* ------------------------------------------------------------------------ */
 void read_superblock(int fd)
 {
+	int x;
+
 	sbd1 = (struct gfs_sb *)&sbd.sd_sb;
 	ioctl(fd, BLKFLSBUF, 0);
 	do_lseek(fd, 0x10 * 4096);
@@ -1451,8 +1444,9 @@ void read_superblock(int fd)
 	sbd.qcsize = GFS2_DEFAULT_QCSIZE;
 	sbd.time = time(NULL);
 	osi_list_init(&sbd.rglist);
-	init_buf_list(&sbd, &sbd.buf_list, 128 << 20);
-	init_buf_list(&sbd, &sbd.nvbuf_list, 0xffffffff);
+	osi_list_init(&sbd.buf_list);
+	for (x = 0; x < BUF_HASH_SIZE; x++)
+		osi_list_init(&sbd.buf_hash[x]);
 	compute_constants(&sbd);
 	gfs2_sb_in(&sbd.sd_sb, buf); /* parse it out into the sb structure */
 	/* Check to see if this is really gfs1 */
@@ -1626,123 +1620,6 @@ uint64_t pop_block(void)
 	return blockstack[bhst].block;
 }
 
-/* ------------------------------------------------------------------------ */
-/* find_journal_block - figure out where a journal starts, given the name   */
-/* Returns: journal block number, changes j_size to the journal size        */
-/* ------------------------------------------------------------------------ */
-uint64_t find_journal_block(const char *journal, uint64_t *j_size)
-{
-	int journal_num;
-	uint64_t jindex_block, jblock = 0;
-	int amtread;
-	struct gfs2_buffer_head *jindex_bh, *j_bh;
-	char jbuf[sbd.bsize];
-	struct gfs2_inode *j_inode = NULL;
-
-	journal_num = atoi(journal + 7);
-	/* Figure out the block of the jindex file */
-	if (gfs1)
-		jindex_block = sbd1->sb_jindex_di.no_addr;
-	else
-		jindex_block = masterblock("jindex");
-	/* read in the block */
-	jindex_bh = bread(&sbd.buf_list, jindex_block);
-	/* get the dinode data from it. */
-	gfs2_dinode_in(&di, jindex_bh->b_data); /* parse disk inode to struct*/
-
-	if (!gfs1)
-		do_dinode_extended(&di, jindex_bh->b_data); /* parse dir. */
-	brelse(jindex_bh, not_updated);
-
-	if (gfs1) {
-		struct gfs2_inode *jiinode;
-		struct gfs_jindex ji;
-
-		jiinode = inode_get(&sbd, jindex_bh);
-		amtread = gfs2_readi(jiinode, (void *)&jbuf,
-				   journal_num * sizeof(struct gfs_jindex),
-				   sizeof(struct gfs_jindex));
-		if (amtread) {
-			gfs_jindex_in(&ji, jbuf);
-			jblock = ji.ji_addr;
-			*j_size = ji.ji_nsegment * 0x10;
-		}
-	} else {
-		struct gfs2_dinode jdi;
-
-		jblock = indirect->ii[0].dirent[journal_num + 2].block;
-		j_bh = bread(&sbd.buf_list, jblock);
-		j_inode = inode_get(&sbd, j_bh);
-		gfs2_dinode_in(&jdi, j_bh->b_data);/* parse dinode to struct */
-		*j_size = jdi.di_size;
-		brelse(j_bh, not_updated);
-	}
-	return jblock;
-}
-
-/* ------------------------------------------------------------------------ */
-/* Check if the word is a keyword such as "sb" or "rindex"                  */
-/* Returns: block number if it is, else 0                                   */
-/* ------------------------------------------------------------------------ */
-uint64_t check_keywords(const char *kword)
-{
-	uint64_t blk = 0;
-
-	if (!strcmp(kword, "sb") ||!strcmp(kword, "superblock"))
-		blk = 0x10 * (4096 / sbd.bsize); /* superblock */
-	else if (!strcmp(kword, "root") || !strcmp(kword, "rootdir"))
-		blk = sbd.sd_sb.sb_root_dir.no_addr;
-	else if (!strcmp(kword, "master")) {
-		if (!gfs1)
-			blk = sbd.sd_sb.sb_master_dir.no_addr;
-		else
-			fprintf(stderr, "This is GFS1; there's no master directory.\n");
-	}
-	else if (!strcmp(kword, "jindex")) {
-		if (gfs1)
-			blk = sbd1->sb_jindex_di.no_addr;
-		else
-			blk = masterblock("jindex"); /* journal index */
-	}
-	else if (!gfs1 && !strcmp(kword, "per_node"))
-		blk = masterblock("per_node");
-	else if (!gfs1 && !strcmp(kword, "inum"))
-		blk = masterblock("inum");
-	else if (!strcmp(kword, "statfs")) {
-		if (gfs1)
-			blk = gfs1_license_di.no_addr;
-		else
-			blk = masterblock("statfs");
-	}
-	else if (!strcmp(kword, "rindex") || !strcmp(kword, "rgindex")) {
-		if (gfs1)
-			blk = sbd1->sb_rindex_di.no_addr;
-		else
-			blk = masterblock("rindex");
-	} else if (!strcmp(kword, "rgs")) {
-		blk = RGLIST_DUMMY_BLOCK;
-	} else if (!strcmp(kword, "quota")) {
-		if (gfs1)
-			blk = gfs1_quota_di.no_addr;
-		else
-			blk = masterblock("quota");
-	} else if (!strncmp(kword, "rg ", 3)) {
-		int rgnum = 0;
-
-		rgnum = atoi(kword + 3);
-		blk = get_rg_addr(rgnum);
-	} else if (!strncmp(kword, "journal", 7) && isdigit(kword[7])) {
-		uint64_t j_size;
-
-		blk = find_journal_block(kword, &j_size);
-	} else if (kword[0]=='0' && kword[1]=='x') /* hex addr */
-		sscanf(kword, "%"SCNx64, &blk);/* retrieve in hex */
-	else
-		sscanf(kword, "%" PRIu64, &blk); /* retrieve decimal */
-
-	return blk;
-}
-
 /* ------------------------------------------------------------------------ */
 /* goto_block - go to a desired block entered by the user                   */
 /* ------------------------------------------------------------------------ */
@@ -1754,8 +1631,32 @@ uint64_t goto_block(void)
 	memset(string, 0, sizeof(string));
 	sprintf(string,"%"PRId64, block);
 	if (bobgets(string, 1, 7, 16, &ch)) {
-		if (isalnum(string[0]))
-			temp_blk = check_keywords(string);
+		if (!strcmp(string,"root"))
+			temp_blk = sbd.sd_sb.sb_root_dir.no_addr;
+		else if (!strcmp(string,"master")) {
+			if (!gfs1)
+				temp_blk = sbd.sd_sb.sb_master_dir.no_addr;
+			else
+				; /* maybe put out an error message at some point */
+		}
+		else if (isalpha(string[0])) {
+			if (gfs1) {
+				if (!strcmp(string, "jindex"))
+					temp_blk = sbd1->sb_jindex_di.no_addr;
+				else if (!strcmp(string, "rindex"))
+					temp_blk = sbd1->sb_rindex_di.no_addr;
+				else if (!strcmp(string, "quota"))
+					temp_blk = gfs1_quota_di.no_addr;
+				else if (!strcmp(string, "rgs"))
+					temp_blk = RGLIST_DUMMY_BLOCK;
+			}
+			else {
+				if (!strcmp(string, "rgs"))
+					temp_blk = RGLIST_DUMMY_BLOCK;
+				else
+					temp_blk = masterblock(string);
+			}
+		}
 		else if (string[0] == '+') {
 			if (string[1] == '0' && string[2] == 'x')
 				sscanf(string, "%"SCNx64, &temp_blk);
@@ -1770,6 +1671,10 @@ uint64_t goto_block(void)
 				sscanf(string, "%" PRIu64, &temp_blk);
 			temp_blk -= block;
 		}
+		else if (string[0] == '0' && string[1] == 'x')
+			sscanf(string, "%"SCNx64, &temp_blk); /* retrieve in hex */
+		else
+			sscanf(string, "%" PRIu64, &temp_blk); /* retrieve decimal */
 
 		if (temp_blk == RGLIST_DUMMY_BLOCK || temp_blk < max_block) {
 			offset = 0;
@@ -2234,130 +2139,17 @@ void gfs_log_header_print(struct gfs_log_header *lh)
 	pv(lh, lh_last_dump, "%llu", "%llx");
 }
 
-/* ------------------------------------------------------------------------ */
-/* print_ld_blocks - print all blocks given in a log descriptor             */
-/* returns: the number of block numbers it printed                          */
-/* ------------------------------------------------------------------------ */
-int print_ld_blocks(const uint64_t *b, const char *end, int start_line)
-{
-	int bcount = 0, i = 0;
-	static char str[256];
-
-	while (*b && (char *)b < end) {
-		if (!termlines ||
-		    (print_entry_ndx >= start_row[dmode] &&
-		     ((print_entry_ndx - start_row[dmode])+1) *
-		     lines_per_row[dmode] <= termlines - start_line - 2)) {
-			if (i && i % 4 == 0) {
-				eol(0);
-				print_gfs2("                    ");
-			}
-			i++;
-			sprintf(str, "0x%llx",
-				(unsigned long long)be64_to_cpu(*b));
-			print_gfs2("%-18.18s ", str);
-			bcount++;
-		}
-		b++;
-		if (gfs1)
-			b++;
-	}
-	eol(0);
-	return bcount;
-}
-
-/* ------------------------------------------------------------------------ */
-/* fsck_readi - same as libgfs2's gfs2_readi, but sets absolute block #     */
-/*              of the first bit of data read.                              */
-/* ------------------------------------------------------------------------ */
-int fsck_readi(struct gfs2_inode *ip, void *buf, uint64_t offset,
-	       unsigned int size, uint64_t *abs_block)
-{
-	struct gfs2_sbd *sdp = ip->i_sbd;
-	struct gfs2_buffer_head *bh;
-	uint64_t lblock, dblock;
-	unsigned int o;
-	uint32_t extlen = 0;
-	unsigned int amount;
-	int not_new = 0;
-	int isdir = !!(S_ISDIR(ip->i_di.di_mode));
-	int copied = 0;
-
-	*abs_block = 0;
-	if (offset >= ip->i_di.di_size)
-		return 0;
-	if ((offset + size) > ip->i_di.di_size)
-		size = ip->i_di.di_size - offset;
-	if (!size)
-		return 0;
-	if (isdir) {
-		lblock = offset;
-		o = do_div(lblock, sdp->sd_jbsize);
-	} else {
-		lblock = offset >> sdp->sd_sb.sb_bsize_shift;
-		o = offset & (sdp->bsize - 1);
-	}
-
-	if (!ip->i_di.di_height) /* inode_is_stuffed */
-		o += sizeof(struct gfs2_dinode);
-	else if (isdir)
-		o += sizeof(struct gfs2_meta_header);
-
-	while (copied < size) {
-		amount = size - copied;
-		if (amount > sdp->bsize - o)
-			amount = sdp->bsize - o;
-		if (!extlen)
-			block_map(ip, lblock, &not_new, &dblock, &extlen,
-				  FALSE, not_updated);
-		if (dblock) {
-			bh = bread(&sdp->buf_list, dblock);
-			if (*abs_block == 0)
-				*abs_block = bh->b_blocknr;
-			dblock++;
-			extlen--;
-		} else
-			bh = NULL;
-		if (bh) {
-			memcpy(buf, bh->b_data + o, amount);
-			brelse(bh, not_updated);
-		} else {
-			memset(buf, 0, amount);
-		}
-		copied += amount;
-		lblock++;
-		o = (isdir) ? sizeof(struct gfs2_meta_header) : 0;
-	}
-	return copied;
-}
-
-void check_journal_wrap(uint64_t seq, uint64_t *highest_seq)
-{
-	if (seq < *highest_seq) {
-		print_gfs2("------------------------------------------------"
-			   "------------------------------------------------");
-		eol(0);
-		print_gfs2("Journal wrapped here.");
-		eol(0);
-		print_gfs2("------------------------------------------------"
-			   "------------------------------------------------");
-		eol(0);
-	}
-	*highest_seq = seq;
-}
-
 /* ------------------------------------------------------------------------ */
 /* dump_journal - dump a journal file's contents.                           */
 /* ------------------------------------------------------------------------ */
 void dump_journal(const char *journal)
 {
-	struct gfs2_buffer_head *j_bh = NULL;
-	uint64_t jblock, j_size, jb, abs_block;
+	struct gfs2_buffer_head *jindex_bh, *j_bh = NULL;
+	uint64_t jindex_block, jblock, j_size, jb;
 	int error, start_line, journal_num;
+	struct gfs2_dinode jdi;
 	char jbuf[sbd.bsize];
 	struct gfs2_inode *j_inode = NULL;
-	int ld_blocks = 0;
-	uint64_t highest_seq = 0;
 
 	start_line = line;
 	lines_per_row[dmode] = 1;
@@ -2365,31 +2157,57 @@ void dump_journal(const char *journal)
 	journal_num = atoi(journal + 7);
 	print_gfs2("Dumping journal #%d.", journal_num);
 	eol(0);
-	jblock = find_journal_block(journal, &j_size);
-	if (!jblock)
-		return;
-	if (!gfs1) {
-		j_bh = bread(&sbd.buf_list, jblock);
+	/* Figure out the block of the jindex file */
+	if (gfs1)
+		jindex_block = sbd1->sb_jindex_di.no_addr;
+	else
+		jindex_block = masterblock("jindex");
+	/* read in the block */
+	jindex_bh = bread(&sbd, jindex_block);
+	/* get the dinode data from it. */
+	gfs2_dinode_in(&di, jindex_bh->b_data); /* parse disk inode into structure */
+
+	if (!gfs1)
+		do_dinode_extended(&di, jindex_bh->b_data); /* parse dir. */
+	brelse(jindex_bh, not_updated);
+
+	if (gfs1) {
+		struct gfs2_inode *jiinode;
+		struct gfs_jindex ji;
+
+		jiinode = inode_get(&sbd, jindex_bh);
+		error = gfs2_readi(jiinode, (void *)&jbuf,
+				   journal_num * sizeof(struct gfs_jindex),
+				   sizeof(struct gfs_jindex));
+		if (!error)
+			return;
+		gfs_jindex_in(&ji, jbuf);
+		jblock = ji.ji_addr;
+		j_size = ji.ji_nsegment * 0x10;
+	} else {
+		jblock = indirect->ii[0].dirent[journal_num + 2].block;
+		j_bh = bread(&sbd, jblock);
 		j_inode = inode_get(&sbd, j_bh);
+		gfs2_dinode_in(&jdi, j_bh->b_data);/* parse dinode to struct */
+		j_size = jdi.di_size;
 	}
 
 	for (jb = 0; jb < j_size; jb += (gfs1 ? 1:sbd.bsize)) {
 		if (gfs1) {
 			if (j_bh)
 				brelse(j_bh, not_updated);
-			j_bh = bread(&sbd.buf_list, jblock + jb);
-			abs_block = jblock + jb;
+			j_bh = bread(&sbd, jblock + jb);
 			memcpy(jbuf, j_bh->b_data, sbd.bsize);
-		} else {
-			error = fsck_readi(j_inode, (void *)&jbuf, jb,
-					   sbd.bsize, &abs_block);
-			if (!error) /* end of file */
-				break;
 		}
+		else
+			error = gfs2_readi(j_inode, (void *)&jbuf, jb,
+					   sbd.bsize);
+		if (!error) /* end of file */
+			break;
 		if (get_block_type(jbuf) == GFS2_METATYPE_LD) {
 			uint64_t *b;
 			struct gfs2_log_descriptor ld;
-			int ltndx;
+			int i = 0, ltndx;
 			uint32_t logtypes[2][6] = {
 				{GFS2_LOG_DESC_METADATA,
 				 GFS2_LOG_DESC_REVOKE,
@@ -2407,8 +2225,8 @@ void dump_journal(const char *journal)
 				{"Metadata", "Unlinked inode", "Dealloc inode",
 				 "Quota", "Final Entry", "Unknown"}};
 
-			print_gfs2("0x%llx (j+%4llx): Log descriptor, ",
-				   abs_block, jb / (gfs1 ? 1 : sbd.bsize));
+			print_gfs2("Block #%4llx: Log descriptor, ",
+				   jb / (gfs1 ? 1 : sbd.bsize));
 			gfs2_log_descriptor_in(&ld, jbuf);
 			print_gfs2("type %d ", ld.ld_type);
 
@@ -2421,50 +2239,47 @@ void dump_journal(const char *journal)
 			print_gfs2("len:%u, data1: %u",
 				   ld.ld_length, ld.ld_data1);
 			eol(0);
-			print_gfs2("                    ");
+			print_gfs2("             ");
 			if (gfs1)
 				b = (uint64_t *)(jbuf +
 					sizeof(struct gfs_log_descriptor));
 			else
 				b = (uint64_t *)(jbuf +
 					sizeof(struct gfs2_log_descriptor));
-			ld_blocks = ld.ld_data1;
-			ld_blocks -= print_ld_blocks(b, (jbuf + sbd.bsize),
-						     start_line);
+			while (*b && (char *)b < (jbuf + sbd.bsize)) {
+				if (!termlines ||
+				    (print_entry_ndx >= start_row[dmode] &&
+				     ((print_entry_ndx - start_row[dmode])+1) *
+				     lines_per_row[dmode] <= termlines - start_line - 2)) {
+					if (i && i % 4 == 0) {
+						eol(0);
+						print_gfs2("             ");
+					}
+					i++;
+					print_gfs2("0x%08llx   ", be64_to_cpu(*b));
+				}
+				b++;
+			}
+			eol(0);
 		} else if (get_block_type(jbuf) == GFS2_METATYPE_LH) {
 			struct gfs2_log_header lh;
 			struct gfs_log_header lh1;
 
 			if (gfs1) {
 				gfs_log_header_in(&lh1, jbuf);
-				check_journal_wrap(lh1.lh_sequence,
-						   &highest_seq);
-				print_gfs2("0x%llx (j+%4llx): Log header: "
-					   "Flags:%x, Seq: 0x%x, "
-					   "1st: 0x%x, tail: 0x%x, "
-					   "last: 0x%x", abs_block,
-					   jb, lh1.lh_flags, lh1.lh_sequence,
-					   lh1.lh_first, lh1.lh_tail,
-					   lh1.lh_last_dump);
+				print_gfs2("Block #%4llx: Log header: Seq"
+					   "= 0x%x, first = 0x%x tail = "
+					   "0x%x, last = 0x%x",
+					   jb, lh1.lh_sequence, lh1.lh_first,
+					   lh1.lh_tail, lh1.lh_last_dump);
 			} else {
 				gfs2_log_header_in(&lh, jbuf);
-				check_journal_wrap(lh.lh_sequence,
-						   &highest_seq);
-				print_gfs2("0x%llx (j+%4llx): Log header: Seq"
-					   ": 0x%x, tail: 0x%x, blk: 0x%x",
-					   abs_block,
+				print_gfs2("Block #%4llx: Log header: Seq"
+					   "= 0x%x, tail = 0x%x, blk = 0x%x",
 					   jb / sbd.bsize, lh.lh_sequence,
 					   lh.lh_tail, lh.lh_blkno);
 			}
 			eol(0);
-		} else if (gfs1 && ld_blocks > 0) {
-			print_gfs2("0x%llx (j+%4llx): GFS log descriptor"
-				   " continuation block", abs_block, jb);
-			eol(0);
-			print_gfs2("                    ");
-			ld_blocks -= print_ld_blocks((uint64_t *)jbuf,
-						     (jbuf + sbd.bsize),
-						     start_line);
 		}
 	}
 	brelse(j_bh, not_updated);
@@ -2502,7 +2317,6 @@ void usage(void)
 	fprintf(stderr,"     rg X - print resource group X.\n");
 	fprintf(stderr,"     rgs - prints all the resource groups (rgs).\n");
 	fprintf(stderr,"     quota - prints the quota file.\n");
-	fprintf(stderr,"-s   specifies a starting block such as root, rindex, quota, inum.\n");
 	fprintf(stderr,"-x   print in hexmode.\n");
 	fprintf(stderr,"-h   prints this help.\n\n");
 	fprintf(stderr,"Examples:\n");
@@ -2580,31 +2394,8 @@ void process_parameters(int argc, char *argv[], int pass)
 				strcpy(device, argv[i]);
 		}
 		else { /* second pass */
-			if (!strcasecmp(argv[i], "-s")) {
-				i++;
-				if (i >= argc - 1) {
-					printf("Error: starting block not specified with -s.\n");
-					printf("%s -s [starting block | keyword] <device>\n",
-					       argv[0]);
-					printf("For example: %s -s \"rg 3\" /dev/exxon_vg/exxon_lv\n",
-					       argv[0]);
-					exit(EXIT_FAILURE);
-				}
-				starting_blk = check_keywords(argv[i]);
-			}
-			else if (!termlines && !strchr(argv[i],'/')) { /* if print, no slash */
-				uint64_t keyword_blk;
-
-				if (!strncmp(argv[i], "journal", 7) &&
-				    isdigit(argv[i][7])) {
-					dump_journal(argv[i]);
-					continue;
-				}
-				keyword_blk = check_keywords(argv[i]);
-				if (keyword_blk) {
-					push_block(keyword_blk);
-				}
-				else if (!strcasecmp(argv[i], "-x"))
+			if (!termlines && !strchr(argv[i],'/')) { /* if print, no slash */
+				if (!strcasecmp(argv[i], "-x"))
 					dmode = HEX_MODE;
 				else if (argv[i][0] == '-') /* if it starts with a dash */
 					; /* ignore it--meant for pass == 0 */
@@ -2613,6 +2404,64 @@ void process_parameters(int argc, char *argv[], int pass)
 				else if (!strcmp(argv[i], "size"))
 					printf("Device size: %" PRIu64 " (0x%" PRIx64 ")\n",
 						   max_block, max_block);
+				else if (!strcmp(argv[i], "sb") ||
+						 !strcmp(argv[i], "superblock"))
+					push_block(0x10 * (4096 / sbd.bsize)); /* superblock */
+				else if (!strcmp(argv[i], "root") ||
+						 !strcmp(argv[i], "rootdir"))
+					push_block(sbd.sd_sb.sb_root_dir.no_addr);
+				else if (!strcmp(argv[i], "master")) {
+					if (gfs1) {
+						fprintf(stderr, "Error: 'master' is invalid "
+							"for GFS (1) file systems.\n");
+						exit(-1);
+					}
+					push_block(sbd.sd_sb.sb_master_dir.no_addr);
+				}
+				else if (!strcmp(argv[i], "jindex")) {
+					if (gfs1)
+						push_block(sbd1->sb_jindex_di.no_addr);
+					else
+						push_block(masterblock("jindex"));/* journal index */
+				}
+				else if (!strcmp(argv[i], "per_node")) {
+					if (gfs1) {
+						fprintf(stderr, "Error: 'per_node' is invalid "
+							"for GFS (1) file systems.\n");
+						exit(-1);
+					}
+					push_block(masterblock("per_node"));
+				}
+				else if (!strcmp(argv[i], "inum")) {
+					if (gfs1) {
+						fprintf(stderr, "Error: 'inum' is invalid "
+							"for GFS (1) file systems.\n");
+						exit(-1);
+					}
+					push_block(masterblock("inum"));
+				}
+				else if (!strcmp(argv[i], "statfs")) {
+					if (gfs1)
+						push_block(gfs1_license_di.no_addr);
+					else
+						push_block(masterblock("statfs"));
+				}
+				else if (!strcmp(argv[i], "rindex") ||
+						 !strcmp(argv[i], "rgindex")) {
+					if (gfs1)
+						push_block(sbd1->sb_rindex_di.no_addr);
+					else
+						push_block(masterblock("rindex"));
+				}
+				else if (!strcmp(argv[i], "rgs")) {
+					push_block(RGLIST_DUMMY_BLOCK);
+				}
+				else if (!strcmp(argv[i], "quota")) {
+					if (gfs1)
+						push_block(gfs1_quota_di.no_addr);
+					else
+						push_block(masterblock("quota"));
+				}
 				else if (!strcmp(argv[i], "rgcount"))
 					rgcount();
 				else if (!strcmp(argv[i], "rgflags")) {
@@ -2670,10 +2519,17 @@ void process_parameters(int argc, char *argv[], int pass)
 					savemeta(argv[i+2], 1);
 				else if (!strcasecmp(argv[i], "savergs"))
 					savemeta(argv[i+2], 2);
+				else if (argv[i][0]=='0' && argv[i][1]=='x') { /* hex addr */
+					sscanf(argv[i], "%"SCNx64, &temp_blk);/* retrieve in hex */
+					push_block(temp_blk);
+				}
 				else if (isdigit(argv[i][0])) { /* decimal addr */
 					sscanf(argv[i], "%"SCNd64, &temp_blk);
 					push_block(temp_blk);
 				}
+				else if (!strncmp(argv[i], "journal", 7) && isdigit(argv[i][7])) {
+					dump_journal(argv[i]);
+				}
 				else {
 					fprintf(stderr,"I don't know what '%s' means.\n", argv[i]);
 					usage();
@@ -2714,7 +2570,7 @@ int main(int argc, char *argv[])
 	dmode = HEX_MODE;
 	sbd.bsize = 4096;
 	type_alloc(buf, char, sbd.bsize); /* allocate/malloc a new 4K buffer */
-	block = starting_blk = 0x10;
+	block = 0x10;
 	for (i = 0; i < BLOCK_STACK_SIZE; i++) {
 		blockstack[i].dmode = dmode;
 		blockstack[i].block = block;
@@ -2740,13 +2596,13 @@ int main(int argc, char *argv[])
 
 	read_superblock(fd);
 	max_block = lseek(fd, 0, SEEK_END) / sbd.bsize;
+	blockstack[0].block = 0x10 * (4096 / sbd.bsize);
 	strcpy(sbd.device_name, device);
 	if (!gfs1)
 		read_master_dir();
 	block_in_mem = -1;
-	process_parameters(argc, argv, 1); /* get what to print from cmdline */
-
-	block = blockstack[0].block = starting_blk * (4096 / sbd.bsize);
+	if (!termlines)    /* if printing to stdout */
+		process_parameters(argc, argv, 1); /* get what to print from cmdline */
 
 	if (termlines)
 		interactive_mode();
