@@ -62,7 +62,7 @@ uint64_t blk_alloc_i(struct gfs2_sbd *sdp, unsigned int type)
 	rg = &rl->rg;
 
 	for (block = 0; block < ri->ri_length; block++) {
-		bh = bread(&sdp->nvbuf_list, ri->ri_addr + block);
+		bh = bread(sdp, ri->ri_addr + block);
 		x = (block) ? sizeof(struct gfs2_meta_header) : sizeof(struct gfs2_rgrp);
 
 		for (; x < sdp->bsize; x++)
@@ -104,7 +104,7 @@ uint64_t blk_alloc_i(struct gfs2_sbd *sdp, unsigned int type)
 
 	brelse(bh, updated);
 
-	bh = bread(&sdp->nvbuf_list, ri->ri_addr);
+	bh = bread(sdp, ri->ri_addr);
 	gfs2_rgrp_out(rg, bh->b_data);
 	brelse(bh, updated);
 
@@ -135,20 +135,20 @@ uint64_t dinode_alloc(struct gfs2_sbd *sdp)
 	return blk_alloc_i(sdp, DINODE);
 }
 
-static __inline__ void buffer_clear_tail(struct gfs2_sbd *sdp,
-					 struct gfs2_buffer_head *bh, int head)
+static __inline__ void buffer_clear_tail(struct gfs2_buffer_head *bh, int head)
 {
-	memset(bh->b_data + head, 0, sdp->bsize - head);
+	memset(bh->b_data + head, 0, bh->b_size - head);
 }
 
 static __inline__ void
-buffer_copy_tail(struct gfs2_sbd *sdp,
-		 struct gfs2_buffer_head *to_bh, int to_head,
-		 struct gfs2_buffer_head *from_bh, int from_head)
+buffer_copy_tail(struct gfs2_buffer_head *to_bh, int to_head,
+				 struct gfs2_buffer_head *from_bh, int from_head)
 {
-	memcpy(to_bh->b_data + to_head, from_bh->b_data + from_head,
-	       sdp->bsize - from_head);
-	memset(to_bh->b_data + sdp->bsize + to_head - from_head, 0,
+	memcpy(to_bh->b_data + to_head,
+	       from_bh->b_data + from_head,
+	       from_bh->b_size - from_head);
+	memset(to_bh->b_data + to_bh->b_size + to_head - from_head,
+	       0,
 	       from_head - to_head);
 }
 
@@ -162,7 +162,7 @@ static void unstuff_dinode(struct gfs2_inode *ip)
 	if (ip->i_di.di_size) {
 		if (isdir) {
 			block = meta_alloc(ip);
-			bh = bget(&sdp->buf_list, block);
+			bh = bget(sdp, block);
 			{
 				struct gfs2_meta_header mh;
 				mh.mh_magic = GFS2_MAGIC;
@@ -171,23 +171,22 @@ static void unstuff_dinode(struct gfs2_inode *ip)
 				gfs2_meta_header_out(&mh, bh->b_data);
 			}
 
-			buffer_copy_tail(sdp, bh,
-					 sizeof(struct gfs2_meta_header),
+			buffer_copy_tail(bh, sizeof(struct gfs2_meta_header),
 					 ip->i_bh, sizeof(struct gfs2_dinode));
 
 			brelse(bh, updated);
 		} else {
 			block = data_alloc(ip);
-			bh = bget(&sdp->buf_list, block);
+			bh = bget(sdp, block);
 
-			buffer_copy_tail(sdp, bh, 0,
+			buffer_copy_tail(bh, 0,
 					 ip->i_bh, sizeof(struct gfs2_dinode));
 
 			brelse(bh, updated);
 		}
 	}
 
-	buffer_clear_tail(sdp, ip->i_bh, sizeof(struct gfs2_dinode));
+	buffer_clear_tail(ip->i_bh, sizeof(struct gfs2_dinode));
 
 	if (ip->i_di.di_size) {
 		*(uint64_t *)(ip->i_bh->b_data + sizeof(struct gfs2_dinode)) = cpu_to_be64(block);
@@ -240,7 +239,7 @@ void build_height(struct gfs2_inode *ip, int height)
 
 		if (new_block) {
 			block = meta_alloc(ip);
-			bh = bget(&sdp->buf_list, block);
+			bh = bget(sdp, block);
 			{
 				struct gfs2_meta_header mh;
 				mh.mh_magic = GFS2_MAGIC;
@@ -248,14 +247,13 @@ void build_height(struct gfs2_inode *ip, int height)
 				mh.mh_format = GFS2_FORMAT_IN;
 				gfs2_meta_header_out(&mh, bh->b_data);
 			}
-			buffer_copy_tail(sdp, bh,
-					 sizeof(struct gfs2_meta_header),
+			buffer_copy_tail(bh, sizeof(struct gfs2_meta_header),
 					 ip->i_bh, sizeof(struct gfs2_dinode));
 
 			brelse(bh, updated);
 		}
 
-		buffer_clear_tail(sdp, ip->i_bh, sizeof(struct gfs2_dinode));
+		buffer_clear_tail(ip->i_bh, sizeof(struct gfs2_dinode));
 
 		if (new_block) {
 			*(uint64_t *)(ip->i_bh->b_data + sizeof(struct gfs2_dinode)) = cpu_to_be64(block);
@@ -266,7 +264,11 @@ void build_height(struct gfs2_inode *ip, int height)
 	}
 }
 
-struct metapath *find_metapath(struct gfs2_inode *ip, uint64_t block)
+struct metapath {
+	unsigned int mp_list[GFS2_MAX_META_HEIGHT];
+};
+
+static struct metapath *find_metapath(struct gfs2_inode *ip, uint64_t block)
 {
 	struct gfs2_sbd *sdp = ip->i_sbd;
 	struct metapath *mp;
@@ -281,6 +283,16 @@ struct metapath *find_metapath(struct gfs2_inode *ip, uint64_t block)
 	return mp;
 }
 
+static __inline__ uint64_t *
+metapointer(struct gfs2_buffer_head *bh, unsigned int height,
+			struct metapath *mp)
+{
+	unsigned int head_size = (height > 0) ?
+		sizeof(struct gfs2_meta_header) : sizeof(struct gfs2_dinode);
+
+	return ((uint64_t *)(bh->b_data + head_size)) + mp->mp_list[height];
+}
+
 static void lookup_block(struct gfs2_inode *ip,
 	     struct gfs2_buffer_head *bh, unsigned int height, struct metapath *mp,
 	     int create, int *new, uint64_t *block)
@@ -359,13 +371,13 @@ void block_map(struct gfs2_inode *ip, uint64_t lblock, int *new,
 
 		if (*new) {
 			struct gfs2_meta_header mh;
-			bh = bget(&sdp->buf_list, *dblock);
+			bh = bget(sdp, *dblock);
 			mh.mh_magic = GFS2_MAGIC;
 			mh.mh_type = GFS2_METATYPE_IN;
 			mh.mh_format = GFS2_FORMAT_IN;
 			gfs2_meta_header_out(&mh, bh->b_data);
 		} else
-			bh = bread(&sdp->buf_list, *dblock);
+			bh = bread(sdp, *dblock);
 	}
 
 	if (!prealloc)
@@ -458,7 +470,7 @@ int gfs2_readi(struct gfs2_inode *ip, void *buf,
 				  FALSE, not_updated);
 
 		if (dblock) {
-			bh = bread(&sdp->buf_list, dblock);
+			bh = bread(sdp, dblock);
 			dblock++;
 			extlen--;
 		} else
@@ -532,7 +544,7 @@ int gfs2_writei(struct gfs2_inode *ip, void *buf,
 		}
 
 		if (new) {
-			bh = bget(&sdp->buf_list, dblock);
+			bh = bget(sdp, dblock);
 			if (isdir) {
 				struct gfs2_meta_header mh;
 				mh.mh_magic = GFS2_MAGIC;
@@ -541,7 +553,7 @@ int gfs2_writei(struct gfs2_inode *ip, void *buf,
 				gfs2_meta_header_out(&mh, bh->b_data);
 			}
 		} else
-			bh = bread(&sdp->buf_list, dblock);
+			bh = bread(sdp, dblock);
 		copy_from_mem(bh, &buf, o, amount);
 		brelse(bh, updated);
 
@@ -578,9 +590,9 @@ struct gfs2_buffer_head *get_file_buf(struct gfs2_inode *ip, uint64_t lbn,
 		ip->i_di.di_size = (lbn + 1) << sdp->sd_sb.sb_bsize_shift;
 
 	if (new)
-		return bget(&sdp->buf_list, dbn);
+		return bget(sdp, dbn);
 	else
-		return bread(&sdp->buf_list, dbn);
+		return bread(sdp, dbn);
 }
 
 int gfs2_dirent_first(struct gfs2_inode *dip, struct gfs2_buffer_head *bh,
@@ -603,7 +615,7 @@ int gfs2_dirent_next(struct gfs2_inode *dip, struct gfs2_buffer_head *bh,
 	char *bh_end;
 	uint16_t cur_rec_len;
 
-	bh_end = bh->b_data + dip->i_sbd->bsize;
+	bh_end = bh->b_data + bh->b_size;
 	cur_rec_len = be16_to_cpu((*dent)->de_rec_len);
 
 	if ((char *)(*dent) + cur_rec_len >= bh_end)
@@ -636,7 +648,7 @@ dirent_alloc(struct gfs2_inode *dip, struct gfs2_buffer_head *bh, int name_len,
 	}
 
 	if (!entries) {
-		dent->de_rec_len = cpu_to_be16(dip->i_sbd->bsize - offset);
+		dent->de_rec_len = cpu_to_be16(bh->b_size - offset);
 		dent->de_name_len = cpu_to_be16(name_len);
 
 		*dent_out = dent;
@@ -734,7 +746,7 @@ dir_split_leaf(struct gfs2_inode *dip, uint32_t index, uint64_t leaf_no)
 	int count;
 
 	bn = meta_alloc(dip);
-	nbh = bget(&dip->i_sbd->buf_list, bn);
+	nbh = bget(dip->i_sbd, bn);
 	{
 		struct gfs2_meta_header mh;
 		mh.mh_magic = GFS2_MAGIC;
@@ -746,7 +758,7 @@ dir_split_leaf(struct gfs2_inode *dip, uint32_t index, uint64_t leaf_no)
 	nleaf = (struct gfs2_leaf *)nbh->b_data;
 	nleaf->lf_dirent_format = cpu_to_be32(GFS2_FORMAT_DE);
 
-	obh = bread(&dip->i_sbd->buf_list, leaf_no);
+	obh = bread(dip->i_sbd, leaf_no);
 	oleaf = (struct gfs2_leaf *)obh->b_data;
 
 	len = 1 << (dip->i_di.di_depth - be16_to_cpu(oleaf->lf_depth));
@@ -877,7 +889,7 @@ int gfs2_get_leaf(struct gfs2_inode *dip, uint64_t leaf_no,
 {
 	int error = 0;
 
-	*bhp = bread(&dip->i_sbd->buf_list, leaf_no);
+	*bhp = bread(dip->i_sbd, leaf_no);
 	if (error)
 		return error;
 	error = gfs2_check_meta(*bhp, GFS2_METATYPE_LF);
@@ -901,7 +913,7 @@ static int get_first_leaf(struct gfs2_inode *dip, uint32_t index,
 	uint64_t leaf_no;
 
 	gfs2_get_leaf_nr(dip, index, &leaf_no);
-	*bh_out = bget(&dip->i_sbd->buf_list, leaf_no);
+	*bh_out = bget(dip->i_sbd, leaf_no);
 	return 0;
 }
 
@@ -923,7 +935,7 @@ static int get_next_leaf(struct gfs2_inode *dip,struct gfs2_buffer_head *bh_in,
 
 	if (!leaf->lf_next)
 		return -1;
-	*bh_out = bget(&dip->i_sbd->buf_list, be64_to_cpu(leaf->lf_next));
+	*bh_out = bget(dip->i_sbd, be64_to_cpu(leaf->lf_next));
 	return 0;
 }
 
@@ -949,7 +961,7 @@ dir_e_add(struct gfs2_inode *dip, char *filename, int len,
 	gfs2_get_leaf_nr(dip, index, &leaf_no);
 
 	for (;;) {
-		bh = bread(&dip->i_sbd->buf_list, leaf_no);
+		bh = bread(dip->i_sbd, leaf_no);
 		leaf = (struct gfs2_leaf *)bh->b_data;
 
 		if (dirent_alloc(dip, bh, len, &dent)) {
@@ -971,7 +983,7 @@ dir_e_add(struct gfs2_inode *dip, char *filename, int len,
 
 			} else {
 				bn = meta_alloc(dip);
-				nbh = bget(&dip->i_sbd->buf_list, bn);
+				nbh = bget(dip->i_sbd, bn);
 				{
 					struct gfs2_meta_header mh;
 					mh.mh_magic = GFS2_MAGIC;
@@ -1022,7 +1034,7 @@ dir_make_exhash(struct gfs2_inode *dip)
 	uint64_t *lp, bn;
 
 	bn = meta_alloc(dip);
-	bh = bget(&sdp->buf_list, bn);
+	bh = bget(sdp, bn);
 	{
 		struct gfs2_meta_header mh;
 		mh.mh_magic = GFS2_MAGIC;
@@ -1035,7 +1047,7 @@ dir_make_exhash(struct gfs2_inode *dip)
 	leaf->lf_dirent_format = cpu_to_be32(GFS2_FORMAT_DE);
 	leaf->lf_entries = cpu_to_be16(dip->i_di.di_entries);
 
-	buffer_copy_tail(sdp, bh, sizeof(struct gfs2_leaf),
+	buffer_copy_tail(bh, sizeof(struct gfs2_leaf),
 			 dip->i_bh, sizeof(struct gfs2_dinode));
 
 	x = 0;
@@ -1053,7 +1065,7 @@ dir_make_exhash(struct gfs2_inode *dip)
 
 	brelse(bh, updated);
 
-	buffer_clear_tail(sdp, dip->i_bh, sizeof(struct gfs2_dinode));
+	buffer_clear_tail(dip->i_bh, sizeof(struct gfs2_dinode));
 
 	lp = (uint64_t *)(dip->i_bh->b_data + sizeof(struct gfs2_dinode));
 
@@ -1106,7 +1118,7 @@ init_dinode(struct gfs2_sbd *sdp, struct gfs2_inum *inum,
 	struct gfs2_buffer_head *bh;
 	struct gfs2_dinode di;
 
-	bh = bget(&sdp->buf_list, inum->no_addr);
+	bh = bget(sdp, inum->no_addr);
 
 	memset(&di, 0, sizeof(struct gfs2_dinode));
 	di.di_header.mh_magic = GFS2_MAGIC;
@@ -1374,7 +1386,7 @@ static int dir_l_search(struct gfs2_inode *dip, const char *filename,
 	if(!inode_is_stuffed(dip))
 		return -1;
 
-	dibh = bread(&dip->i_sbd->buf_list, dip->i_di.di_num.no_addr);
+	dibh = bread(dip->i_sbd, dip->i_di.di_num.no_addr);
 	error = leaf_search(dip, dibh, filename, len, &dent, NULL);
 	if (!error) {
 		gfs2_inum_in(inum, (char *)&dent->de_inum);
@@ -1429,7 +1441,7 @@ static int dir_e_del(struct gfs2_inode *dip, const char *filename, int len)
 		gfs2_get_leaf_nr(dip, index, &leaf_no);
 
 		while(leaf_no && !found){
-			bh = bget(&dip->i_sbd->buf_list, leaf_no);
+			bh = bget(dip->i_sbd, leaf_no);
 			error = leaf_search(dip, bh, filename, len, &cur, &prev);
 			if (error) {
 				if(error != -ENOENT){
@@ -1463,7 +1475,7 @@ static int dir_l_del(struct gfs2_inode *dip, struct gfs2_buffer_head *dibh,
 		return -1;
 
 	if(!dibh) {
-		dibh = bread(&dip->i_sbd->buf_list, dip->i_di.di_num.no_addr);
+		dibh = bread(dip->i_sbd, dip->i_di.di_num.no_addr);
 		if (error)
 			return -1;
 		got_buf = 1;
@@ -1546,23 +1558,6 @@ int gfs2_lookupi(struct gfs2_inode *dip, const char *filename, int len,
 	return error;
 }
 
-/**
- * gfs2_free_block - free up a block given its block number
- */
-void gfs2_free_block(struct gfs2_sbd *sdp, uint64_t block)
-{
-	struct gfs2_buffer_head *bh;
-	struct rgrp_list *rgd;
-
-	gfs2_set_bitmap(sdp, block, GFS2_BLKST_FREE);
-	/* Adjust the free space count for the freed block */
-	rgd = gfs2_blk2rgrpd(sdp, block); /* find the rg for indir block */
-	bh = bget(&sdp->nvbuf_list, rgd->ri.ri_addr); /* get the rg buffer */
-	rgd->rg.rg_free++; /* adjust the free count */
-	gfs2_rgrp_out(&rgd->rg, bh->b_data); /* back to the buffer */
-	brelse(bh, updated); /* release the buffer */
-}
-
 /**
  * gfs2_freedi - unlink a disk inode by block number.
  * Note: currently only works for regular files.
@@ -1576,7 +1571,7 @@ int gfs2_freedi(struct gfs2_sbd *sdp, uint64_t block)
 	unsigned char *buf;
 	struct rgrp_list *rgd;
 	
-	bh = bread(&sdp->buf_list, block);
+	bh = bread(sdp, block);
 	ip = inode_get(sdp, bh);
 	if (ip->i_di.di_height > 0) {
 		buf = (unsigned char *)bh->b_data;
@@ -1584,8 +1579,16 @@ int gfs2_freedi(struct gfs2_sbd *sdp, uint64_t block)
 		for (x = sizeof(struct gfs2_dinode); x < sdp->bsize;
 			 x += sizeof(uint64_t)) {
 			p = be64_to_cpu(*(uint64_t *)(buf + x));
-			if (p)
-				gfs2_free_block(sdp, p);
+			if (p) {
+				gfs2_set_bitmap(sdp, p, GFS2_BLKST_FREE);
+				/* We need to adjust the free space count for the freed */
+                /* indirect block. */
+				rgd = gfs2_blk2rgrpd(sdp, p); /* find the rg for indir block */
+				bh = bget(sdp, rgd->ri.ri_addr); /* get the buffer its rg */
+				rgd->rg.rg_free++; /* adjust the free count */
+				gfs2_rgrp_out(&rgd->rg, bh->b_data); /* back to the buffer */
+				brelse(bh, updated); /* release the buffer */
+			}
 		}
 	}
 	/* Set the bitmap type for inode to free space: */
@@ -1596,7 +1599,7 @@ int gfs2_freedi(struct gfs2_sbd *sdp, uint64_t block)
 	/* The rg itself is in memory as rgd->rg, but there's most likely a  */
 	/* buffer in memory for the rg on disk because we used it to fix the */
 	/* bitmaps, some of which are on the same block on disk.             */
-	bh = bread(&sdp->nvbuf_list, rgd->ri.ri_addr); /* get the buffer */
+	bh = bread(sdp, rgd->ri.ri_addr); /* get the buffer */
 	rgd->rg.rg_free++;
 	rgd->rg.rg_dinodes--; /* one less inode in use */
 	gfs2_rgrp_out(&rgd->rg, bh->b_data);
