@@ -90,17 +90,6 @@ meta_data()
 	    <content type="boolean"/>
 	</parameter>
 
-	<parameter name="sleeptime">
-	    <longdesc lang="en">
-		Amount of time to sleep after removing an IP address.
-		Value is specified in seconds. Default value is 10.
-	    </longdesc>
-	    <shortdesc lang="en">
-		Amount of time (seconds) to sleep.
-	    </shortdesc>
-	    <content type="string"/>
-	</parameter>
-
     </parameters>
 
     <actions>
@@ -533,22 +522,13 @@ network_link_up()
 	#
 	if [ "${intf_arg/bond/}" != "$intf_arg" ]; then
 		
-		#
-		# With Xen we must check for slaves of pbondX, not bondX
-		#
-		masterif=$intf_arg
-		line=$(ip link list | awk {'print $2'} | grep "p$intf_arg:")
-		if [ $? -eq 0 ] ; then
-			masterif="p$intf_arg"
-		fi
-		
 		#
 		# Bonded driver.  Check link of all slaves for this interface.
 		# If any link is up, the bonding driver is expected to route
 		# traffic through that link.  Thus, the entire bonded link
 		# is declared up.
 		#
-		slaves=$(findSlaves $masterif)
+		slaves=$(findSlaves $intf_arg)
 		if [ $? -ne 0 ]; then
 			ocf_log err "Error finding slaves of $intf_arg"
 			return 1
@@ -694,7 +674,7 @@ ipv4()
 			if [ "${addr/\/*/}" != "$ifaddr" ]; then
 			        continue
 			fi
-			addr=`/sbin/ip addr list | grep "$addr/" | head -n 1 | awk '{print $2}'`
+			addr=`/sbin/ip addr list | grep "$addr" | head -n 1 | awk '{print $2}'`
 			ocf_log info "Removing IPv4 address $addr from $dev"
 		fi
 		
@@ -907,13 +887,7 @@ stop)
 
 		# XXX Let nfsd/lockd clear their queues; we hope to have a
 		# way to enforce this in the future
-		if [ -z "$OCF_RESKEY_sleeptime" ]; then
-		    sleep 10
-		else
-		    if [ "$OCF_RESKEY_sleeptime" -gt "0" ]; then
-			sleep $OCF_RESKEY_sleeptime
-		    fi
-		fi
+		sleep 10
 	else
 		ocf_log debug "${OCF_RESKEY_address} is not configured"
 	fi
