@@ -108,27 +108,6 @@ meta_data()
 	    </shortdesc>
 	    <content type="boolean" default="0"/>
 	</parameter>
-
-	<parameter name="nfs_client_cache">
-            <longdesc lang="en">
-	   	On systems with large numbers of exports, a performance
-		problem in the exportfs command can cause inordinately long
-		status check times for services with lots of mounted
-		NFS clients.  This occurs because exportfs does DNS queries
-		on all clients in the export list.
-
-		Setting this option to '1' will enable caching of the export
-		list returned from the exportfs command on a per-service
-		basis.  The cache will last for 30 seconds before expiring
-		instead of being generated each time an nfsclient resource
-		is called.
-            </longdesc>
-            <shortdesc lang="en">
-	    	Enable exportfs list caching
-            </shortdesc>
-	    <content type="integer" default="0"/>
-	</parameter>
-
                 
         <parameter name="recovery" reconfig="1">
             <longdesc lang="en">
@@ -215,7 +194,7 @@ meta_data()
     </actions>
     
     <special tag="rgmanager">
-        <attributes maxinstances="1"/>
+        <attributes root="1" maxinstances="1"/>
         <child type="lvm" start="1" stop="9"/>
         <child type="fs" start="2" stop="8"/>
         <child type="clusterfs" start="3" stop="7"/>
